import numpy as np
import pandas as pd
import pytest
from matplotlib import pyplot as plt

from pyhrp.cluster import Asset, Portfolio


def test_portfolio():
    p = Portfolio()
    a = Asset(name="A")
    b = Asset(name="B")
    c = Asset(name="C")

    p[a] = 0.4
    p[b] = 0.3
    p[c] = 0.3

    cov = pd.DataFrame(index=[a, b, c], columns=[a, b, c], data=np.diag([2, 3, 4]))

    ax = p.plot(names=["A", "B", "C"])
    ax.set_facecolor("lightyellow")  # Change the background color
    ax.grid(True)  # Enable gridlines

    plt.show()

    assert p.variance(cov) == pytest.approx(0.95)


def test_getset_item():
    p = Portfolio()
    a = Asset(name="A")
    p[a] = 0.4
    assert p[a] == 0.4


def test_asset_equality():
    a = Asset(name="A")
    b = Asset(name="A")
    assert a == b
    assert not a < 2



import matplotlib.pyplot as plt

from pyhrp.algos import one_over_n
from pyhrp.hrp import build_tree


def test_one_over_n(returns):
    cor = returns.corr()
    dendrogram = build_tree(cor=cor, method="ward")
    dendrogram.plot()
    plt.show()

    for level, portfolio in one_over_n(dendrogram):
        print(f"Level: {level}")
        print(portfolio)



import pytest

from pyhrp.algos import one_over_n
from pyhrp.cluster import Asset
from pyhrp.cluster import Cluster as Node
from pyhrp.hrp import Dendrogram


def test_one_over_n():
    root = Node(10)
    root.left = Node(11)
    root.right = Node(0)
    a = Asset(name="A")
    # root.right.asset = a

    root.left.left = Node(1)
    b = Asset(name="B")
    # root.left.left.asset = b

    root.left.right = Node(2)
    c = Asset(name="C")
    # root.left.right.asset = c
    print(root)

    dendrogram = Dendrogram(root=root, assets=[a, b, c])

    for level, portfolio in one_over_n(dendrogram):
        print(f"Level: {level}")
        print(portfolio)


def test_wrong_number_of_nodes():
    root = Node(10)
    root.left = Node(11)
    root.right = Node(0)
    root.left.left = Node(1)
    root.left.right = Node(2)

    a = Asset(name="A")
    b = Asset(name="B")
    c = Asset(name="C")
    d = Asset(name="D")

    with pytest.raises(ValueError):
        Dendrogram(root=root, assets=[a, b, c, d])



import marimo

__generated_with = "0.11.5"
app = marimo.App()


@app.cell
def _(mo):
    mo.md(
        r"""
        # 1 over N (the hierarchical version)

        Inspired by Thomas Raffinot

        - Compute a dendrogram using the 'ward' distance method.
          Do not compute a 2nd Dendrogram.
        - Apply the methods level by level. Go from level 0 (only the root),
          to level 1 (left and right of the root), to level 2 (...)
        - On level n evaluate the function f for all leaves for each node.

        """
    )
    return


@app.cell
def _(__file__):
    from pathlib import Path

    data = Path(__file__).parent / "data"
    return Path, data


@app.cell
def _(data):
    import pandas as pd

    from pyhrp.cluster import Asset

    prices = pd.read_csv(data / "stock_prices.csv", index_col=0)
    returns = prices.pct_change().dropna(axis=0, how="all").fillna(0.0)
    returns.columns = [Asset(name=column) for column in returns.columns]
    return Asset, pd, prices, returns


@app.cell
def _(returns):
    cor = returns.corr()
    cov = returns.cov()
    return cor, cov


@app.cell
def _():
    import matplotlib.pyplot as plt

    from pyhrp.hrp import build_tree

    return build_tree, plt


@app.cell
def _(build_tree, cor, plt):
    # We first build the tree. This task is very separated from the computations of weights on such a tree.
    dendrogram = build_tree(cor, method="ward")
    dendrogram.plot()
    plt.show()
    return (dendrogram,)


@app.cell
def _(dendrogram, plt):
    # We use the tree from the previous step and perform a 1/n
    # strategy in an iterative manner
    from pyhrp.algos import one_over_n

    # Drill deeper, level by level
    # Root is level 0 at Level 1 are two nodes...
    for level, portfolio in one_over_n(dendrogram):
        print(f"Level: {level}")
        portfolio.plot(names=dendrogram.names)
        plt.show()
    return level, one_over_n, portfolio


@app.cell
def _():
    import marimo as mo

    return (mo,)


if __name__ == "__main__":
    app.run()



import numpy as np
import pandas as pd
from pypfopt import HRPOpt

from pyhrp.hrp import hrp


def test_allocation(prices):
    # reproduce the results of the implementation in PyPortfolioOpt
    returns = prices.pct_change().dropna(axis=0, how="all")

    optimizer = HRPOpt(returns)
    weights = pd.Series(optimizer.optimize(linkage_method="single"))
    print(weights)

    cluster = hrp(prices=prices, method="single", bisection=True)
    w = cluster.portfolio.weights
    print(w)

    assert np.linalg.norm(weights - w[weights.index]) < 0.006



from __future__ import annotations

from copy import deepcopy
from typing import Generator

import pandas as pd

from .cluster import Cluster, Portfolio


def risk_parity(root: Cluster, cov: pd.DataFrame) -> Cluster:
    """compute a cluster"""
    if root.is_leaf:
        # a node is a leaf if has no further relatives downstream.
        asset = cov.keys().to_list()[root.value]
        root.portfolio[asset] = 1.0
        return root

    # drill down on the left
    root.left = risk_parity(root.left, cov)
    # drill down on the right
    root.right = risk_parity(root.right, cov)

    # combine left and right into a new cluster
    return _parity(root, cov=cov)


def _parity(cluster, cov) -> Cluster:
    """
    Given two clusters compute in a bottom-up approach their parent.

    :param cluster: left cluster
    :param cov: covariance matrix. Will pick the correct sub-matrix

    """
    # split is s.t. v_left * alpha_left == v_right * alpha_right and alpha + beta = 1
    v_left = cluster.left.portfolio.variance(cov)
    v_right = cluster.right.portfolio.variance(cov)

    alpha_left, alpha_right = v_right / (v_left + v_right), v_left / (v_left + v_right)

    # assets in the cluster are the assets of the left and right cluster
    # further downstream
    assets = {
        **(alpha_left * cluster.left.portfolio.weights).to_dict(),
        **(alpha_right * cluster.right.portfolio.weights).to_dict(),
    }

    for asset, weight in assets.items():
        cluster.portfolio[asset] = weight

    return cluster


def one_over_n(dendrogram) -> Generator[tuple[int, Portfolio]]:
    root = dendrogram.root
    assets = dendrogram.assets

    w = 1

    for n, level in enumerate(root.levels):
        for node in level:
            for leaf in node.leaves:
                root.portfolio[assets[leaf.value]] = w / node.leaf_count

        w *= 0.5
        yield n, deepcopy(root.portfolio)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

import numpy as np
import pandas as pd
from binarytree import Node


@dataclass(frozen=True)
class Asset:
    mu: float = None
    name: str = None

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Asset):
            return False
        return self.name == other.name

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, Asset):
            return False
        return self.name < other.name


@dataclass
class Portfolio:
    # _variance: float = None
    _weights: dict[Asset, float] = field(default_factory=dict)

    @property
    def assets(self) -> list[Asset]:
        return list(self._weights.keys())

    def variance(self, cov: pd.DataFrame) -> float:
        c = cov[self.assets].loc[self.assets].values
        w = self.weights[self.assets].values
        return np.linalg.multi_dot((w, c, w))

    def __getitem__(self, item) -> float:
        return self._weights[item]

    def __setitem__(self, key, value: float):
        self._weights[key] = value

    @property
    def weights(self):
        """weight series"""
        return pd.Series(self._weights, name="Weights").sort_index()

    # @property
    def weights_vs_name(self, names):
        w = {asset.name: weight for asset, weight in self.weights.items()}
        return pd.Series({name: w[name] for name in names})

    def plot(self, names):
        # Plot the weights using pandas' built-in plotting, without needing to import matplotlib
        # a = self.weights_vs_name(names)
        a = self.weights_vs_name(names)

        ax = a.plot(kind="bar", color="skyblue")

        # Set x-axis labels and rotations
        ax.set_xticklabels(names, rotation=90, fontsize=8)
        return ax


class Cluster(Node):
    """
    Clusters are the nodes of the graphs we build.
    Each cluster is aware of the left and the right cluster
    it is connecting to.
    """

    def __init__(self, value, left: Cluster | None = None, right: Cluster | None = None, **kwargs):
        super().__init__(value=value, left=left, right=right, **kwargs)
        self.portfolio = Portfolio()

    @property
    def is_leaf(self):
        return self.left is None and self.right is None

    @property
    def leaves(self):
        """
        Give a set of all reachable leaf nodes.

        Note that the leaves method of the Node class implemented in BinaryTree
        is not respecting the 'correct' order of the nodes.
        """
        if self.is_leaf:
            return [self]
        else:
            return self.left.leaves + self.right.leaves
